// -- file extension rules --

/** Extensions that must have EXIF/etc data stripped */
export const extScrubExif = new Set([
  ".jpg",
  ".jpeg",
  ".png",
  ".mov",
  ".mp4",
  ".m4a",
]);
/** Extensions that rendered syntax-highlighted code */
export const extsCode = new Map<string, highlight.Language>(Object.entries({
  ".json": "json",
  ".toml": "toml",
  ".ts": "ts",
  ".js": "ts",
  ".tsx": "tsx",
  ".jsx": "tsx",
  ".css": "css",
  ".py": "python",
  ".lua": "lua",
  ".sh": "shell",
  ".bat": "dosbatch",
  ".ps1": "powershell",
  ".cmd": "dosbatch",
  ".yaml": "yaml",
  ".yml": "yaml",
  ".zig": "zig",
  ".astro": "astro",
  ".mdx": "mdx",
  ".xml": "xml",
  ".jsonc": "json",
  ".php": "php",
  ".patch": "diff",
  ".diff": "diff",
}));
/** These files show an audio embed. */
export const extsAudio = new Set([
  ".mp3",
  ".flac",
  ".wav",
  ".ogg",
  ".m4a",
]);
/** These files show a video embed. */
export const extsVideo = new Set([
  ".mp4",
  ".mkv",
  ".webm",
  ".avi",
  ".mov",
]);
/** These files show an image embed */
export const extsImage = new Set([
  ".jpg",
  ".jpeg",
  ".png",
  ".webp",
  ".avif",
  ".heic",
]);
/** These files show an image embed, but aren't optimized */
export const extsImageLike = new Set([
  ...extsImage,
  ".svg",
  ".gif",
]);

/** These files populate `duration` using `ffprobe` */
export const extsDuration = new Set([...extsAudio, ...extsVideo]);
/** These files populate `dimensions` using `ffprobe` */
export const extsDimensions = new Set([...extsImage, ...extsVideo]);

/** These files read file contents into `contents`, as-is */
export const extsReadContents = new Set([".txt", ".chat", ".dirsort"]);

export const extsArchive = new Set([
  ".zip",
  ".rar",
  ".7z",
  ".tar",
  ".gz",
  ".bz2",
  ".xz",
]);

/**
 * Formats which are already compression formats, meaning a pass
 * through zstd would offer little to negative benefits
 */
export const extsPreCompressed = new Set([
  ...extsAudio,
  ...extsVideo,
  ...extsImageLike,
  ...extsArchive,
  ".docx",
  ".xlsx",
  ".pptx",
  ".psd",
  ".sketch",
  ".ai",
  ".3ds",
  ".fbx",
  ".blend",
  ".dng",
]);
extsPreCompressed.delete(".svg");

export function fileIcon(
  file: Pick<MediaFile, "kind" | "basename" | "path">,
  dirOpen?: boolean,
) {
  const { kind, basename } = file;
  if (kind === MediaFileKind.directory) return dirOpen ? "dir-open" : "dir";

  // -- special cases --
  if (file.path === "/2024/for everyone") return "snow";

  // -- basename cases --
  if (basename === "readme.txt") return "readme";

  // -- extension cases --
  const ext = path.extname(file.basename).toLowerCase();
  if ([".comp", ".fuse", ".setting"].includes(ext)) return "fusion";
  if ([".json", ".toml", ".yaml", ".yml"].includes(ext)) return "json";
  if (ext === ".blend") return "blend";
  if (ext === ".chat") return "chat";
  if (ext === ".html") return "webpage";
  if (ext === ".lnk") return "link";
  if (ext === ".txt" || ext === ".md") return "text";

  // -- extension categories --
  if (extsVideo.has(ext)) return "video";
  if (extsAudio.has(ext)) return "audio";
  if (extsImage.has(ext)) return "image";
  if (extsArchive.has(ext)) return "archive";
  if (extsCode.has(ext)) return "code";

  return "file";
}

// -- viewer rules --
const pathToCanvas = new Map<string, string>(Object.entries({
  "/2017": "2017",
  "/2018": "2018",
  "/2019": "2019",
  "/2020": "2020",
  "/2021": "2021",
  "/2022": "2022",
  "/2023": "2023",
  "/2024": "2024",
}));

import type * as highlight from "./highlight.ts";
import { MediaFile, MediaFileKind } from "@/file-viewer/models/MediaFile.ts";
import * as path from "node:path";
