// This canvas is based on the maze generation algo in Tanks. This was
// originally written in C++ as a single function in 2018, and was ported to TS
// by Chloe in 2025 for the cotyledon canvas.
//
// The main difference is that this version is a visualization, rather than the
// practical function. Instead of taking a millisecond, only 5 steps are
// performed per second, visualizing the whole ordeal. It also isn't a playable
// game, obviously.
//
// Ported with love because I care about my old self
// She deserves the world, but instead gave it to me.
(globalThis as any).canvas_2018 = function (canvas: HTMLCanvasElement) {
  const isStandalone = canvas.getAttribute("data-standalone") === "true";
  if (isStandalone) {
    canvas.style.backgroundColor = "#27201E";
  }
  interface Cell {
    down: boolean;
    right: boolean;
    visited: boolean;

    cell_flash: number;
    down_flash: number;
    right_flash: number;
  }
  interface Pos {
    x: number;
    y: number;
    /** Where the wall is relative to x, y. */
    dir: "left" | "right" | "up" | "down";
  }
  interface Maze {
    grid: Grid;
    cursor: { x: number; y: number };
    lastTick: number;
    /* Pixels */
    transform: number;
    newCellsToVisit: Pos[];
    randomWallBag: Cell[];
    randomWallTarget: number;
    renderOffset: { x: number; y: number };
    done: boolean;
  }
  const hex = (color: number[]) =>
    "#" + color.map((c) => c.toString(16).padStart(2, "0")).join("");
  let cellSize: number;
  let borderThickness: number;
  const cellFlashModifier = isStandalone ? 0.4 : 0.2;
  const color = isStandalone ? "#170d0b" : "#231C1A";
  const bg = [0x27, 0x20, 0x1E];
  const wallFlashColor = [0xFF, 0xA8, 0x7A];
  const cellFlashColor = "#FFA87A";
  const updateTime = 1000 / 7;
  const randomWallBreakInterval = [6, 12]; // every 10 to 18 walls.
  function randomBetween(min: number, max: number) {
    return Math.round(
      Math.random() * (max - min),
    ) + min;
  }
  function randomOf<T>(array: T[]): T {
    return array[randomBetween(0, array.length - 1)];
  }
  function randomWallTarget() {
    return randomBetween(
      randomWallBreakInterval[0],
      randomWallBreakInterval[1],
    );
  }

  // Originally, this used a 2-dimensional array. However, I wanted to make sure
  // that the grid could be infinitely sized. This grid constructs new cells on
  // demand, as needed.
  class Grid {
    cells = new Map<number, Cell>();
    cell({ x, y }: { x: number; y: number }) {
      const k = ((x | 0) << 16) + (y | 0);
      const { cells } = this;
      let existing = this.cells.get(k);
      if (!existing) {
        existing = {
          cell_flash: 0,
          down: true,
          down_flash: 0,
          right: true,
          right_flash: 0,
          visited: false,
        };
        cells.set(k, existing);
      }
      return existing;
    }
    forAll(
      renderOffset: { x: number; y: number },
      width: number,
      height: number,
      cb: (cell: Cell, pos: { x: number; y: number }) => void,
    ) {
      const { x: offsetX, y: offsetY } = renderOffset;
      const startX = Math.floor(-offsetX / cellSize);
      const startY = Math.floor(-offsetY / cellSize);
      const endX = Math.ceil((width - offsetX) / cellSize);
      const endY = Math.ceil((height - offsetY) / cellSize);
      for (let x = startX; x <= endX; x++) {
        for (let y = startY; y <= endY; y++) {
          const cellX = offsetX + x * cellSize;
          const cellY = offsetY + y * cellSize;
          cb(this.cell({ x, y }), { x: cellX, y: cellY });
        }
      }
    }
  }

  const ctx = canvas.getContext("2d")!;
  if (!ctx) {
    console.error("Could not get canvas context");
    return () => {};
  }

  let width: number, height: number;
  const updateDimensions = () => {
    width = canvas.width = canvas.offsetWidth;
    height = canvas.height = canvas.offsetHeight;
    cellSize = 100;
    borderThickness = 8;
  };
  updateDimensions();

  setTimeout(() => {
    updateDimensions();
  }, 10);

  let maze = initMaze();
  let nextMaze: Maze | null = null;
  let completeFade = 0;
  function initMaze(): Maze {
    return {
      grid: new Grid(),
      transform: 0,
      cursor: {
        x: randomBetween(0, Math.ceil(width / cellSize)),
        y: randomBetween(0, Math.ceil(height / cellSize)),
      },
      lastTick: performance.now(),
      randomWallBag: [],
      randomWallTarget: randomWallTarget(),
      newCellsToVisit: [],
      renderOffset: { x: 0, y: 0 },
      done: false,
    };
  }

  function isOnScreen(maze: Maze, x: number, y: number) {
    const { x: offsetX, y: offsetY } = maze.renderOffset;
    const cellX = offsetX + x * cellSize;
    const cellY = offsetY + y * cellSize;
    return (
      cellX + cellSize > 0 &&
      cellX < width &&
      cellY + cellSize > 0 &&
      cellY < height
    );
  }

  function tick(maze: Maze, other?: Maze) {
    if (maze.done) return;

    // The original maze algorithm broke down 4%-8% of random right facing
    // walls, and 4%-8% of down facing walls. It did this at the end.
    // To make this visual more interesting, two random walls will be broken
    // down every 12-25 cell visits. This way, the main trail is always running.
    if (maze.randomWallBag.length > maze.randomWallTarget) {
      const down: Cell = randomOf(maze.randomWallBag);
      const right: Cell = randomOf(maze.randomWallBag);
      maze.randomWallBag.forEach((cell) =>
        cell.cell_flash = Math.min(cell.cell_flash + 0.2, 1)
      );
      down.cell_flash = 1;
      down.down = false;
      down.down_flash = 1;
      right.cell_flash = 1;
      right.right = false;
      right.right_flash = 1;
      maze.randomWallBag = [];
      maze.randomWallTarget = randomWallTarget();
      return;
    }

    // The main algorithm was simple: Have a cursor position, and move it in a
    // random direction that it had not seen before. Once it had run out of
    // options, branch off of a previous location. Only visit each cell once.
    //
    // In this visualization, cells that are too far offscreen are softly
    // treated as "visited", which is how the simulation always stays in frame.
    const current = maze.grid.cell(maze.cursor);
    current.visited = true;
    current.cell_flash = 1;
    maze.randomWallBag.push(current);
    const adjacent = ([
      { x: maze.cursor.x + 1, y: maze.cursor.y, dir: "left" },
      { x: maze.cursor.x - 1, y: maze.cursor.y, dir: "right" },
      { x: maze.cursor.x, y: maze.cursor.y + 1, dir: "up" },
      { x: maze.cursor.x, y: maze.cursor.y - 1, dir: "down" },
    ] as Pos[]).filter((pos) =>
      isOnScreen(maze, pos.x, pos.y) &&
      maze.grid.cell(pos).visited === false
    );
    if (adjacent.length === 0) {
      // move cursor to a random cell that has not been visited.
      const cells = maze.newCellsToVisit.filter((pos) =>
        isOnScreen(maze, pos.x, pos.y) &&
        maze.grid.cell(pos).visited === false
      );
      if (cells.length === 0) {
        maze.done = true;
        return;
      }
      const continuePos = randomOf(cells);
      breakWall(maze, continuePos, other);
      maze.cursor = { x: continuePos.x, y: continuePos.y };
      return;
    }

    // break a random wall
    const toBreak = randomOf(adjacent);
    breakWall(maze, toBreak, other);
    maze.cursor = { x: toBreak.x, y: toBreak.y };

    // add the other directions to the new cells to visit.
    maze.newCellsToVisit.push(
      ...adjacent.filter((pos) => pos.dir !== toBreak.dir),
    );
  }

  function breakWall(maze: Maze, pos: Pos, other?: Maze) {
    if (pos.dir === "right") {
      const cell = maze.grid.cell(pos);
      cell.right = false;
      cell.right_flash = 1;
      if (other) cell.right = false;
    } else if (pos.dir === "down") {
      const cell = maze.grid.cell(pos);
      cell.down = false;
      cell.down_flash = 1;
      if (other) cell.down = false;
    } else if (pos.dir === "left") {
      const cell = maze.grid.cell({ x: pos.x - 1, y: pos.y });
      cell.right = false;
      cell.right_flash = 1;
      if (other) cell.right = false;
    } else if (pos.dir === "up") {
      const cell = maze.grid.cell({ x: pos.x, y: pos.y - 1 });
      cell.down = false;
      cell.down_flash = 1;
      if (other) cell.down = false;
    }
  }

  function renderOffset(maze: Maze) {
    return { x: maze.transform, y: maze.transform };
  }

  let animationFrameId: number;
  let last = performance.now();
  let dt: number = 0;

  function renderMazeBorders(maze: Maze, opacity: number) {
    ctx.globalAlpha = opacity;
    maze.grid.forAll(
      maze.renderOffset,
      width,
      height,
      (cell, { x: cellX, y: cellY }) => {
        // Walls
        if (cell.right) {
          ctx.fillStyle = color;
          ctx.fillRect(
            cellX + cellSize - borderThickness / 2,
            cellY - borderThickness / 2,
            borderThickness,
            cellSize + borderThickness,
          );
        }
        if (cell.down) {
          ctx.fillStyle = color;
          ctx.fillRect(
            cellX - borderThickness / 2,
            cellY + cellSize - borderThickness / 2,
            cellSize + borderThickness,
            borderThickness,
          );
        }
      },
    );
    ctx.globalAlpha = 1;
  }

  function renderCellFlash(maze: Maze) {
    maze.grid.forAll(
      maze.renderOffset,
      width,
      height,
      (cell, { x: cellX, y: cellY }) => {
        // Cell flash to show visiting path.
        if (cell.cell_flash > 0) {
          cell.cell_flash = Math.max(0, cell.cell_flash - dt / 1000);
          ctx.fillStyle = cellFlashColor;
          ctx.globalAlpha = cell.cell_flash * cellFlashModifier;
          ctx.fillRect(cellX, cellY, cellSize, cellSize);
          ctx.globalAlpha = 1;
        }
      },
    );
  }

  function renderBorderFlash(maze: Maze) {
    maze.grid.forAll(
      maze.renderOffset,
      width,
      height,
      (cell, { x: cellX, y: cellY }) => {
        if (cell.right_flash == 0 && cell.down_flash == 0) {
          return;
        }

        // Walls
        const cellFlash = cell.cell_flash * cellFlashModifier;
        if (cell.right_flash > 0) {
          cell.right_flash = Math.max(0, cell.right_flash - dt / 500);
          ctx.fillStyle = interpolateColor(
            bg,
            wallFlashColor,
            Math.max(cell.right_flash, cellFlash),
          );
          if (cellFlash > cell.right_flash) {
            ctx.globalAlpha = cell.right_flash / cellFlash;
          }
          ctx.fillRect(
            cellX + cellSize - borderThickness / 2,
            cellY + borderThickness / 2,
            borderThickness,
            cellSize - borderThickness,
          );
          ctx.globalAlpha = 1;
        }
        if (cell.down_flash > 0) {
          if (cellFlash > cell.down_flash) {
            ctx.globalAlpha = cell.down_flash / cellFlash;
          }
          cell.down_flash = Math.max(0, cell.down_flash - dt / 500);
          ctx.fillStyle = interpolateColor(
            bg,
            wallFlashColor,
            Math.max(cell.down_flash, cellFlash),
          );
          ctx.fillRect(
            cellX + borderThickness / 2,
            cellY + cellSize - borderThickness / 2,
            cellSize - borderThickness,
            borderThickness,
          );
          ctx.globalAlpha = 1;
        }
      },
    );
  }

  function render() {
    const now = performance.now();
    dt = now - last;
    maze.transform += dt * 0.005;
    maze.renderOffset = renderOffset(maze);
    if (!maze.done) {
      if (now - maze.lastTick >= updateTime) {
        tick(maze);
        maze.lastTick = now;

        if (maze.done) {
          nextMaze = initMaze();
          nextMaze.transform = (maze.transform % cellSize) - dt * 0.005;
          nextMaze.lastTick = now;
          completeFade = 0;
        }
      }
    }
    if (nextMaze) {
      nextMaze.transform += dt * 0.005;
      nextMaze.renderOffset = renderOffset(nextMaze);
      if (!nextMaze.done && now - nextMaze.lastTick >= updateTime) {
        tick(nextMaze, maze);
        nextMaze.lastTick = now;
      }
    }
    last = now;

    ctx.clearRect(0, 0, width, height);

    renderCellFlash(maze);
    if (nextMaze) renderCellFlash(nextMaze);

    renderMazeBorders(maze, 1);
    if (nextMaze) {
      renderMazeBorders(nextMaze, completeFade);
      completeFade += dt / 3000;
      if (completeFade >= 1) {
        maze = nextMaze;
        nextMaze = null;
      }
    }

    renderBorderFlash(maze);
    if (nextMaze) {
      renderCellFlash(nextMaze);
      renderBorderFlash(nextMaze);
    }

    animationFrameId = requestAnimationFrame(render);
  }

  function interpolateColor(start: number[], end: number[], t: number) {
    return hex(start.map((s, i) => Math.round(s + (end[i] - s) * t)));
  }

  window.addEventListener("resize", updateDimensions);
  animationFrameId = requestAnimationFrame(render);

  // cleanup function
  return () => {
    window.removeEventListener("resize", updateDimensions);
    cancelAnimationFrame(animationFrameId);
  };
};
